from kivy.uix.screenmanager import Screen
from kivy.uix.popup import Popup
from kivy.properties import StringProperty
from clease.gui.help_message_popup import HelpMessagePopup
from kivy.app import App
from clease.gui.load_save_dialog import LoadDialog


class MCMainPage(Screen):
    _pop_up = None
    mc_cell_lengths = StringProperty('a: b: c:')
    angle_info = StringProperty('alpha: beta: gamma:')
    num_atoms = StringProperty('No. atoms')
    eci_file = StringProperty('')
    mc_cell_size = StringProperty('1')

    def dismiss_popup(self):
        self._pop_up.dismiss()
        self._pop_up = None

    @property
    def main_mc_screen(self):
        app = App.get_running_app()
        return app.root.ids.sm.get_screen('MCHeader')

    @property
    def info_update_disabled(self):
        app = App.get_running_app()
        return app.root.info_update_disabled

    def show_size_help_message(self):
        msg = 'Scaling factor used to scale the currently active template\n'
        msg += 'If this factor is 5 a supercell consiting (5, 5, 5)\n'
        msg += 'extension of the active template will be used for the\n'
        msg += 'MC calculation\n'
        content = HelpMessagePopup(close=self.dismiss_popup, msg=msg)
        self._pop_up = Popup(title="Concentration input", content=content,
                             pos_hint={'right': 0.95, 'top': 0.95},
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def get_mc_cell(self):
        app = App.get_running_app()
        return app.root.get_mc_cell()

    def set_cell_info(self):
        if self.info_update_disabled:
            return

        atoms = self.get_mc_cell()
        if atoms is None:
            return

        info = atoms.get_cell_lengths_and_angles()

        length_str = 'a: {}Å b: {}Å c: {}Å'.format(
            int(info[0]), int(info[1]), int(info[2]))
        self.mc_cell_lengths = length_str

        angle_str = '{}deg {}deg {}deg'.format(int(info[3]), int(info[4]),
                                               int(info[5]))
        self.angle_info = angle_str
        self.num_atoms = 'No. atoms {}'.format(len(atoms))

    def show_load_dialog(self):
        content = LoadDialog(load=self.load, cancel=self.dismiss_popup)
        self._pop_up = Popup(title="Load ECIs", content=content,
                             pos_hint={'right': 0.95, 'top': 0.95},
                             size_hint=(0.9, 0.9))
        self._pop_up.open()

    def load(self, path, filename):
        if len(filename) == 0:
            self.eci_file = path
        else:
            self.eci_file = filename[0]
        self.dismiss_popup()

    def to_dict(self):
        return {
            'size': self.ids.sizeInput.text,
            'eci_file': self.ids.eciFileInput.text
        }

    def from_dict(self, dct):
        self.ids.sizeInput.text = dct.get('size', '1')
        self.ids.eciFileInput.text = dct.get('eci_file', '')



# distutils: language = c++

from libcpp.string cimport string
from libcpp.vector cimport vector
from libcpp.map cimport map as map_cpp

cdef extern from "ce_updater.hpp":
  cdef cppclass CEUpdater:
      CEUpdater() except +

      # Initialize the object
      void init(object atoms, object BC, object corrFunc, object eci, object cluster_info) except +

      # Clear update history
      void clear_history()

      # Undo all chages since last call to clear_history
      void undo_changes()

      # Update the correlation functions
      void update_cf(object system_changes) except +

      double get_energy()

      double calculate(object system_changes) except +

      object get_cf()

      void set_eci(object eci)

      object get_singlets()

      const vector[string]& get_symbols() const

      void set_num_threads(unsigned int num_threads)

      void get_changes(vector[string] &symbs, vector[unsigned int] &changed_sites) except +

      void calculate_cf_from_scratch(const vector[string] &cf_names, map_cpp[string, double] &cf)

      void set_atoms(object atoms) except+



from clease.montecarlo.observers import MCObserver
from clease import _logger
import numpy as np


class LowestEnergyStructure(MCObserver):
    """Track the lowest energy state visited during an MC run.

    atoms: Atoms object
        Atoms object used in Monte Carlo

    verbose: bool
        If `True`, progress messages will be printed
    """

    def __init__(self, atoms, verbose=True):
        self.atoms = atoms
        self.lowest_energy = np.inf
        self.lowest_energy_cf = None
        self.emin_atoms = None

        self.name = "LowestEnergyStructure"
        self.verbose = verbose

    def reset(self):
        self.lowest_energy = np.inf
        self.lowest_energy_cf = None
        self.emin_atoms = None

    def __call__(self, system_changes):
        """
        Check if the current state has lower energy and store the current
        state if it has a lower energy than the previous state.

        system_changes: list
            System changes. See doc-string of
            `clease.montecarlo.observers.MCObserver`
        """
        energy = self.atoms.get_calculator().results['energy']
        if self.emin_atoms is None:
            calc = self.atoms.get_calculator()
            self.lowest_energy_cf = calc.get_cf()
            self.lowest_energy = energy
            self.emin_atoms = self.atoms.copy()
            return

        if energy < self.lowest_energy:
            dE = energy - self.lowest_energy
            calc = self.atoms.get_calculator()
            self.lowest_energy = energy
            self.emin_atoms = self.atoms.copy()
            self.lowest_energy_cf = calc.get_cf()
            if self.verbose:
                msg = "Found new low energy structure. "
                msg += "New energy: {} eV. ".format(self.lowest_energy)
                msg += "Change: {} eV".format(dE)
                _logger(msg)



import json
import os
from clease import CorrFunction
from clease.calculator import Clease
from clease.tools import nested_list2str


def attach_calculator(setting=None, atoms=None, eci={}):
    """Utility function for an efficient initialization of large cells.

    Parameters:

    setting: `ClusterExpansionSetting` object

    eci: dict
        Dictionary containing cluster names and their ECI values

    atoms: Atoms object
        Atoms object for MC simulations.

    """
    cf_names = list(eci.keys())
    init_cf = CorrFunction(setting).get_cf_by_names(setting.atoms, cf_names)

    template = setting.template_atoms.get_template_matching_atoms(
        atoms=atoms)
    setting.set_active_template(atoms=template)

    atoms = setting.atoms.copy()

    calc = Clease(setting, eci=eci, init_cf=init_cf)
    atoms.set_calculator(calc)
    return atoms


def save_info(fname, data):
    with open(fname, 'w') as outfile:
        json.dump(data, outfile)



from clease.cluster_generator import ClusterGenerator
from clease.cluster_list import ClusterList
from clease.cluster import Cluster
from itertools import product, chain
from clease.name_clusters import name_clusters, size
from scipy.spatial import KDTree
from copy import deepcopy
import numpy as np
from ase.geometry import wrap_positions
from clease.cluster_fingerprint import ClusterFingerprint


class ClusterManager(object):
    """
    Manager for construction of all cluster.

    Parameters:

    prim_cell: ase.Atoms
        Primitive cell
    """
    def __init__(self, prim_cell):
        self.generator = ClusterGenerator(prim_cell)
        self.clusters = ClusterList()

    def build(self, max_size=4, max_cluster_dia=4.0):
        """
        Construct all clusters.

        Parameters:

        max_size: int
            Maximum number of atoms in a cluster

        max_cluster_dia: float
            Maximum distance between two atoms in a cluster
        """
        if isinstance(max_cluster_dia, float):
            max_cluster_dia = [max_cluster_dia for _ in range(max_size+1)]
        num_lattices = range(len(self.generator.prim))
        cluster_size = range(2, max_size+1)
        all_fps = []
        names = []
        all_clusters = []
        all_eq_sites = []
        lattices = []
        diameters = []
        sizes = []
        for l, s in product(num_lattices, cluster_size):
            clusters, fps = self.generator.generate(
                s, max_cluster_dia[s], ref_lattice=l)

            eq_sites = []
            for c in clusters:
                eq_sites.append(self.generator.equivalent_sites(c[0]))

            all_fps += fps
            all_clusters += clusters
            all_eq_sites += eq_sites
            lattices += [l]*len(clusters)
            sizes += [s]*len(clusters)
            diameters += [2*np.sqrt(fp[0]) for fp in fps]

        names = self._get_names(all_fps)
        # Transfer to the cluster list
        zipped = zip(names, sizes, diameters, all_fps, all_clusters,
                     all_eq_sites, lattices)
        for n, s, d, fp, c, eq, l in zipped:
            cluster = Cluster(name=n, size=s, diameter=d, fingerprint=fp,
                              ref_indx=-1, indices=c, equiv_sites=eq,
                              trans_symm_group=l)
            self.clusters.append(cluster)

        # Add singlets and empty
        for i in range(len(self.generator.prim)):
            self.clusters.append(Cluster(name='c1', size=1, diameter=0.0,
                                 fingerprint=ClusterFingerprint([1.0]),
                                 ref_indx=-1, indices=[[[0, 0, 0, i]]],
                                 equiv_sites=[], trans_symm_group=i))
            self.clusters.append(Cluster('c0', size=0, diameter=0.0,
                                         fingerprint=ClusterFingerprint([0.0]),
                                         ref_indx=-1, indices=[],
                                         equiv_sites=[], trans_symm_group=i))

    def _get_names(self, all_fps):
        """
        Give a consistent name to all clusters

        Parameter:

        all_fps: list of ClusterFingerPrint
            A list with all the cluster fingerprints
        """
        # The following one-liner will do the job when we
        # decide to update cluster names from the legacy ones
        # where names where per by size
        # names = name_clusters(all_fps)

        sizes = set([size(fp) for fp in all_fps])
        names = [None for _ in all_fps]
        for s in sizes:
            fps = []
            indices = []
            for i, fp in enumerate(all_fps):
                if size(fp) == s:
                    indices.append(i)
                    fps.append(fp)
            names_per_size = name_clusters(fps)
            for i, n in zip(indices, names_per_size):
                names[i] = n
        return names

    def _ref_indices(self, kdtree):
        """
        Return all reference indices

        Parameters:

        kdtree: KDTree or cKDTree
            A KDtree representation of all atomic positions
        """
        ref_indices = []
        for i in range(len(self.generator.prim)):
            d, i = kdtree.query(self.generator.cartesian([0, 0, 0, i]))
            ref_indices.append(i)
        return ref_indices

    def info_for_template(self, template):
        """
        Specialise the cluster information to a template

        Parameter:

        template: ase.Atoms
            Atoms object representing the simulation cell
        """
        unique = self.unique_four_vectors()
        lut = self.fourvec_to_indx(template, unique)
        ref_indices = [lut[(0, 0, 0, i)]
                       for i in range(self.generator.num_sub_lattices)]

        cluster_int = deepcopy(self.clusters)
        for cluster in cluster_int:
            if cluster.size == 0:
                cluster.ref_indx = int(ref_indices[cluster.group])
                cluster.indices = []
            elif cluster.size == 1:
                cluster.ref_indx = int(ref_indices[cluster.group])
                cluster.indices = []
            else:
                cluster.indices = self.generator.to_atom_index(
                    cluster.indices, lut)
                cluster.ref_indx = int(ref_indices[cluster.group])
        return cluster_int

    def unique_four_vectors(self):
        """
        Return a list with all unique 4-vectors consituting
        the clusters
        """
        unique = set()
        for c in self.clusters:
            if c.size == 0:
                continue
            flat = list(chain(*c.indices))
            indices = list(map(tuple, flat))
            unique.update(indices)
        return unique

    def get_figures(self):
        """
        Return a list of atoms object representing the clusters
        """
        return self.clusters.get_figures(self.generator)

    def create_four_vector_lut(self, template):
        """
        Construct a lookup table (LUT) for the index in template given the wrapped
        vector

        Parameter:

        template: Atoms
            Atoms object to use when creating the lookup table (LUT)
        """
        lut = {}
        pos = template.get_positions().copy()
        for i in range(pos.shape[0]):
            if template[i].tag >= self.generator.num_sub_lattices:
                continue
            vec = self.generator.get_four_vector(pos[i, :], template[i].tag)
            lut[tuple(vec)] = i
        return lut

    def fourvec_to_indx(self, template, unique):
        cell = template.get_cell()
        pos = np.zeros((len(unique), 3))
        for i, u in enumerate(unique):
            cart_u = self.generator.cartesian(u)
            pos[i, :] = cart_u

        pos = wrap_positions(pos, cell)
        unique_indices = []
        for i in range(pos.shape[0]):
            diff_sq = np.sum((pos[i, :] - template.get_positions())**2, axis=1)
            unique_indices.append(np.argmin(diff_sq))
        return dict(zip(unique, unique_indices))

    def translation_matrix(self, template):
        """
        Construct the translation matrix

        Parameter:

        template: ase.Atoms
            Atoms object representing the simulation cell
        """
        trans_mat = []
        unique = self.unique_four_vectors()
        cell = template.get_cell()

        unique_indx = self.fourvec_to_indx(template, unique)

        vec2 = np.zeros(4, dtype=int)
        lut = self.create_four_vector_lut(template)
        indices = [0 for _ in range(len(unique))]
        cartesian = np.zeros((len(unique), 3))

        # Make a copy of the positions to avoid atoms being translated
        tmp_pos = template.get_positions().copy()
        for atom in template:
            if atom.tag >= self.generator.num_sub_lattices:
                trans_mat.append({})
                continue

            vec = self.generator.get_four_vector(tmp_pos[atom.index, :], atom.tag)
            for i, u in enumerate(unique):
                vec2[:3] = vec[:3] + u[:3]
                vec2[3] = u[3]

                cartesian[i, :] = self.generator.cartesian(vec2)

            cartesian = wrap_positions(cartesian, cell)

            for i, u in enumerate(unique):
                vec2_four = self.generator.get_four_vector(
                    cartesian[i, :], u[3])
                indices[i] = lut[tuple(vec2_four)]

            trans_mat.append(dict(zip([int(unique_indx[u]) for u in unique],
                                      indices)))
        return trans_mat



from clease.montecarlo.constraints import MCConstraint
import numpy as np


class ConstrainElementInserts(MCConstraint):
    """
    Constrain inserting the elements by basis. This constraint is intended
    to be used together with SGCMonteCarlo

    atoms: Atoms object
        ASE Atoms object used in the MC simulation

    index_by_basis: list
        Indices ordered by basis (same as `index_by_basis` parameter in
        `ClusterExpansionSetting`).
        If an Atoms object has 10 sites where the first 4 belongs to the
        first basis, the next 3 belongs to the next basis and the last 3
        belongs to the last basis, the `index_by_basis` would be
        [[0, 1, 2, 3], [4, 5, 6], [7, 8, 9]]

    element_by_basis: list
        List specifying which elements are allowed in each basis. If there are
        two basis where Si and O are allowed in the fist basis while Si and C
        are allowed in the second basis, the argument would be
        [['Si', 'O'], ['Si', 'C']]
    """
    def __init__(self, atoms, index_by_basis, element_by_basis):
        self.basis = np.zeros(len(atoms), dtype=int)
        for i, indices in enumerate(index_by_basis):
            for x in indices:
                self.basis[x] = i

        num_basis = len(element_by_basis)
        unique_elem = set([x for item in element_by_basis for x in item])
        self.element_allowed = {x: np.zeros(num_basis, dtype=np.uint8)
                                for x in unique_elem}

        # Initialize the element lookup
        for i, elements in enumerate(element_by_basis):
            for elem in elements:
                self.element_allowed[elem][i] = 1

    def __call__(self, system_changes):
        for change in system_changes:
            indx = change[0]
            new_elem = change[2]
            basis = self.basis[indx]
            if self.element_allowed[new_elem][basis] == 0:
                return False
        return True



"""Definition of ClusterExpansionSetting Class.

This module defines the base-class for storing the settings for performing
Cluster Expansion in different conditions.
"""
from copy import deepcopy
import numpy as np
from ase.db import connect

from clease.tools import (wrap_and_sort_by_position,
                          nested_list2str)
from clease.basis_function import BasisFunction
from clease.template_atoms import TemplateAtoms
from clease.concentration import Concentration
from clease import AtomsManager
from clease.cluster_list import ClusterList
from clease.template_filters import ValidConcentrationFilter
from clease.cluster_manager import ClusterManager


class ClusterExpansionSetting(object):
    """Base class for all Cluster Expansion settings."""

    def __init__(self, prim, concentration, size=None, supercell_factor=27,
                 db_name='clease.db', max_cluster_size=4,
                 max_cluster_dia=[5.0, 5.0, 5.0], basis_function='polynomial',
                 skew_threshold=40, ignore_background_atoms=False):
        self.kwargs = {'size': size,
                       'supercell_factor': supercell_factor,
                       'db_name': db_name,
                       'max_cluster_size': max_cluster_size,
                       'max_cluster_dia': deepcopy(max_cluster_dia),
                       'ignore_background_atoms': ignore_background_atoms}

        if isinstance(concentration, Concentration):
            self.concentration = concentration
        elif isinstance(concentration, dict):
            self.concentration = Concentration.from_dict(concentration)
        else:
            raise TypeError("concentration has to be either dict or "
                            "instance of Concentration")

        self.kwargs["concentration"] = self.concentration.to_dict()
        self.cluster_list = ClusterList()
        self.basis_elements = deepcopy(self.concentration.basis_elements)
        self.num_basis = len(self.basis_elements)
        self._check_first_elements()
        self.ignore_background_atoms = ignore_background_atoms
        self.db_name = db_name
        self.size = to_3x3_matrix(size)

        self.prim_cell = prim
        self._tag_prim_cell()
        self._store_prim_cell()

        if self.ignore_background_atoms:
            self.cluster_manager = ClusterManager(self.prim_no_bkg())
        else:
            self.cluster_manager = ClusterManager(self.prim_cell)

        prim_mng = AtomsManager(prim)
        prim_ind_by_basis = prim_mng.index_by_symbol(
            [x[0] for x in self.basis_elements])
        conc_filter = ValidConcentrationFilter(concentration,
                                               prim_ind_by_basis)

        self.template_atoms = TemplateAtoms(
            self.prim_cell, supercell_factor=supercell_factor,
            size=self.size, skew_threshold=skew_threshold,
            filters=[conc_filter])
        self.atoms_mng = AtomsManager(None)

        self.max_cluster_size = max_cluster_size
        self.max_cluster_dia = self._format_max_cluster_dia(max_cluster_dia)
        self.cluster_manager.build(
            max_size=self.max_cluster_size,
            max_cluster_dia=self.max_cluster_dia
        )
        self.all_elements = sorted([item for row in self.basis_elements for
                                    item in row])
        self.background_indices = None
        self.num_elements = len(self.all_elements)
        self.unique_elements = sorted(list(set(deepcopy(self.all_elements))))
        self.num_unique_elements = len(self.unique_elements)
        self.index_by_basis = None

        self.index_by_sublattice = []
        self.ref_index_trans_symm = []
        self.template_atoms_uid = 0

        self.set_active_template(
            atoms=self.template_atoms.weighted_random_template())

        unique_element_no_bkg = self.unique_element_without_background()
        if isinstance(basis_function, BasisFunction):
            if basis_function.unique_elements != unique_element_no_bkg:
                raise ValueError("Unique elements in BasiFunction instance "
                                 "is different from the one in settings")
            self.bf_scheme = basis_function

        elif isinstance(basis_function, str):
            if basis_function.lower() == 'polynomial':
                from clease.basis_function import Polynomial
                self.bf_scheme = Polynomial(unique_element_no_bkg)
            elif basis_function.lower() == 'trigonometric':
                from clease.basis_function import Trigonometric
                self.bf_scheme = Trigonometric(unique_element_no_bkg)
            elif basis_function.lower() == "binary_linear":
                from clease.basis_function import BinaryLinear
                self.bf_scheme = BinaryLinear(unique_element_no_bkg)
            else:
                msg = "basis function scheme {} ".format(basis_function)
                msg += "is not supported."
                raise ValueError(msg)
        else:
            raise ValueError("basis_function has to be instance of "
                             "BasisFunction or a string")

        self.spin_dict = self.bf_scheme.spin_dict
        self.basis_functions = self.bf_scheme.basis_functions

        if len(self.basis_elements) != self.num_basis:
            raise ValueError("list of elements is needed for each basis")

    @property
    def atoms(self):
        return self.atoms_mng.atoms

    def prim_no_bkg(self):
        """
        Return an instance of the primitive cell where the background indices
        has been removed
        """
        prim = self.prim_cell.copy()
        bg_syms = self.get_bg_syms()
        delete = []
        for atom in prim:
            if atom.symbol in bg_syms:
                delete.append(atom.index)

        delete.sort(reverse=True)
        for i in delete:
            del prim[i]
        return prim

    def get_bg_syms(self):
        """
        Return the symbols in the basis where there is only one element
        """
        return set(x[0] for x in self.basis_elements if len(x) == 1)

    def unique_element_without_background(self):
        """Remove backgound elements."""
        if not self.ignore_background_atoms:
            bg_sym = set()
        else:
            bg_sym = self.get_bg_syms()

            # Remove bg_syms that are also present in basis with more than one
            # element
            for elems in self.basis_elements:
                if len(elems) == 1:
                    continue
                to_be_removed = set()
                for s in bg_sym:
                    if s in elems:
                        to_be_removed.add(s)

                bg_sym -= to_be_removed

        unique_elem = set()
        for x in self.basis_elements:
            unique_elem.update(x)
        return list(unique_elem - bg_sym)

    def _size2string(self):
        """Convert the current size into a string."""
        return nested_list2str(self.size)

    def prepare_new_active_template(self, template):
        """Prepare necessary data structures when setting new template."""
        self.template_atoms_uid = 0
        self.size = template.info['size']
        self.atoms_mng.atoms = wrap_and_sort_by_position(template)

        self.index_by_basis = self._group_index_by_basis()

        self.background_indices = self._get_background_indices()
        self.index_by_sublattice = self.atoms_mng.index_by_tag()
        self.num_trans_symm = len(self.index_by_sublattice)
        self.ref_index_trans_symm = [i[0] for i in self.index_by_sublattice]

    def set_active_template(self, size=None, atoms=None):
        """Set a new template atoms object."""
        if size is not None and atoms is not None:
            raise ValueError("Specify either size or pass Atoms object.")
        if size is not None:
            template = self.template_atoms.get_template_with_given_size(
                size=size)
        elif atoms is not None:
            template = self.template_atoms.get_template_matching_atoms(
                atoms=atoms)
        else:
            template = self.template_atoms.weighted_random_template()

        template = wrap_and_sort_by_position(template)

        if atoms is not None:
            # Check that the positions of the generated template
            # matches the ones in the passed object
            atoms = wrap_and_sort_by_position(atoms)
            if not np.allclose(template.get_positions(),
                               atoms.get_positions()):
                raise ValueError("Inconsistent positions. Passed object\n"
                                 "{}\nGenerated template\n{}"
                                 "".format(atoms.get_positions(),
                                           template.get_positions()))

        self.prepare_new_active_template(template)
        self.create_cluster_list_and_trans_matrix()

    def _tag_prim_cell(self):
        """
        Add a tag to all the atoms in the unit cell to track the sublattice.
        Tags are added such that that the lowest tags corresponds to "active"
        sites, while the highest tags corresponds to background sites. An
        example is a system having three sublattices that can be occupied by
        more than one species, and two sublattices that can be occupied by
        only one species. The tags 0, 1, 2 will then be assigned to the three
        sublattices that can be occupied by more than one species, and the two
        remaining lattices will get the tag 3, 4.
        """
        bg_sym = self.get_bg_syms()
        tag = 0

        # Tag non-background elements first
        for atom in self.prim_cell:
            if atom.symbol not in bg_sym:
                atom.tag = tag
                tag += 1

        # Tag background elements
        for atom in self.prim_cell:
            if atom.symbol in bg_sym:
                atom.tag = tag
                tag += 1

    def _store_prim_cell(self):
        """Store unit cell to the database."""
        db = connect(self.db_name)
        shape = self.prim_cell.get_cell_lengths_and_angles()
        for row in db.select(name='primitive_cell'):
            uc_shape = row.toatoms().get_cell_lengths_and_angles()
            if np.allclose(shape, uc_shape):
                return row.id

        uid = db.write(self.prim_cell, name='primitive_cell')
        return uid

    def _get_prim_cell(self):
        raise NotImplementedError("This function has to be implemented in "
                                  "in derived classes.")

    def _format_max_cluster_dia(self, max_cluster_dia):
        """Get max_cluster_dia in numpy array form"""
        # max_cluster_dia is list or array
        if isinstance(max_cluster_dia, (list, np.ndarray)):
            # Length should be either max_cluster_size+1 or max_cluster_size-1
            if len(max_cluster_dia) == self.max_cluster_size + 1:
                for i in range(2):
                    max_cluster_dia[i] = 0.
                max_cluster_dia = np.array(max_cluster_dia, dtype=float)
            elif len(max_cluster_dia) == self.max_cluster_size - 1:
                max_cluster_dia = np.array(max_cluster_dia, dtype=float)
                max_cluster_dia = np.insert(max_cluster_dia, 0, [0., 0.])
            else:
                raise ValueError("Invalid length for max_cluster_dia.")
        # max_cluster_dia is int or float
        elif isinstance(max_cluster_dia, (int, float)):
            max_cluster_dia *= np.ones(self.max_cluster_size - 1, dtype=float)
            max_cluster_dia = np.insert(max_cluster_dia, 0, [0., 0.])
        # Case for *None* or something else
        else:
            raise ValueError("max_cluster_dia must be float, int or list.")
        return max_cluster_dia.round(decimals=3)

    def _get_background_indices(self):
        """Get indices of the background atoms."""
        # check if any basis consists of only one element type
        basis = [i for i, b in enumerate(self.basis_elements) if len(b) == 1]

        bkg_indices = []
        for b_indx in basis:
            bkg_indices += self.index_by_basis[b_indx]
        return bkg_indices

    def _get_atoms(self):
        """Create atoms with a user-specified size."""
        atoms = self.prim_cell.copy() * self.size
        return wrap_and_sort_by_position(atoms)

    @property
    def multiplicity_factor(self):
        """Return the multiplicity factor of each cluster."""
        num_sites_in_group = [len(x) for x in self.index_by_sublattice]
        return self.cluster_list.multiplicity_factors(num_sites_in_group)

    @property
    def all_cf_names(self):
        num_bf = len(self.basis_functions)
        return self.cluster_list.get_all_cf_names(num_bf)

    @property
    def num_cf(self):
        """Return the number of correlation functions."""
        return len(self.all_cf_names)

    def create_cluster_list_and_trans_matrix(self):
        at_cpy = self.atoms
        self.cluster_list = self.cluster_manager.info_for_template(at_cpy)
        self.trans_matrix = self.cluster_manager.translation_matrix(at_cpy)

    def _keys2int(self, tm):
        """
        Convert the keys in the translation matrix to integers
        """
        tm_int = []
        for row in tm:
            tm_int.append({int(k): v for k, v in row.items()})
        return tm_int

    def _group_index_by_basis(self):
        first_symb_in_basis = [x[0] for x in self.basis_elements]
        return self.atoms_mng.index_by_symbol(first_symb_in_basis)

    def view_clusters(self):
        """Display all clusters along with their names."""
        from ase.gui.gui import GUI
        from ase.gui.images import Images
        figures = self.cluster_manager.get_figures()
        images = Images()
        images.initialize(figures)
        gui = GUI(images, expr='')
        gui.show_name = True
        gui.run()

    def get_all_templates(self):
        """
        Return a list with all template atoms
        """
        return self.template_atoms.get_all_templates()

    def view_templates(self):
        """
        Display all templates in the ASE GUi
        """
        from ase.visualize import view
        view(self.get_all_templates())

    def _check_first_elements(self):
        basis_elements = self.basis_elements
        num_basis = self.num_basis
        # This condition can be relaxed in the future
        first_elements = []
        for elements in basis_elements:
            first_elements.append(elements[0])
        if len(set(first_elements)) != num_basis:
            raise ValueError("First element of different basis should not be "
                             "the same.")

    def save(self, filename):
        """Write Setting object to a file in JSON format.

        Parameters:

        filename: str
            Name of the file to store the necessary settings to initialize
            the Cluster Expansion calculations.
        """

        import json
        with open(filename, 'w') as outfile:
            json.dump(self.kwargs, outfile, indent=2)


def to_3x3_matrix(size):
    if size is None:
        return None
    if isinstance(size, np.ndarray):
        size = size.tolist()

    # Is already a matrix
    if np.array(size).shape == (3, 3):
        return size

    if np.array(size).shape == (3,):
        return np.diag(size).tolist()

    raise ValueError("Cannot convert passed array {} to 3x3 matrix"
                     "".format(size))



from itertools import product, permutations, combinations
import numpy as np


class AtomsFilter(object):
    """
    Base class for all template filters that requires a full Atoms object in
    order to decide the validity of the template.
    """

    def __call__(self, atoms):
        """
        The call method has to be implemented in derived classes.
        It accepts an atoms object or a cell object and returns `True` if the
        templates is valid and False if the templates is not valid.
        """
        raise NotImplementedError("Has to be implemented in derived classes.")


class CellFilter(object):
    """
    Base class for all template filters that only require knowledge of the
    cell vectors in order to decide if the template is valid or not.
    """

    def __call__(self, cell):
        raise NotImplementedError("Has to be implemented in derived classes.")


class SkewnessFilter(CellFilter):
    """
    A filter that rejects cells that are too skewed.

    Parameters:

    ratio: float
        Maximum accepted ratio between the longest and the shortest diagonal
    """

    def __init__(self, ratio):
        self.ratio = ratio

    def _get_max_min_diag_ratio(self, cell):
        """Return the ratio between the maximum and the minimum diagonal."""
        diag_lengths = []
        for w in product([-1, 0, 1], repeat=3):
            if np.allclose(w, 0):
                continue
            diag = np.array(w).dot(cell)
            length = np.sqrt(diag.dot(diag))
            diag_lengths.append(length)
        max_length = np.max(diag_lengths)
        min_length = np.min(diag_lengths)
        return max_length/min_length

    def __call__(self, cell):
        diag_ratio = self._get_max_min_diag_ratio(cell)
        return diag_ratio < self.ratio


class EquivalentCellsFilter(CellFilter):
    """
    A filter that rejects templates that can be obtained by a unitary
    transformation of an existing cell.
    """

    def __init__(self, cell_list):
        self.cell_list = cell_list

    def _is_unitary(self, matrix):
        return np.allclose(matrix.T.dot(matrix), np.identity(matrix.shape[0]))

    def _are_equivalent(self, cell1, cell2):
        """Compare two cells to check if they are equivalent.

        It is assumed that the cell vectors are columns of each matrix.
        """
        inv_cell1 = np.linalg.inv(cell1)
        for perm in permutations(range(3)):
            permute_cell = cell2[:, perm]
            R = permute_cell.dot(inv_cell1)
            if self._is_unitary(R):
                return True
        return False

    def __call__(self, cell):
        for existing_cell in self.cell_list:
            if self._are_equivalent(existing_cell, cell):
                return False
        return True


class ValidConcentrationFilter(AtomsFilter):
    """
    A filter that rejects template that has no valid concentration.

    Parameters:

    setting: ClusterExpansionSetting
        Instance of `ClusterExpansionSetting`
    """

    def __init__(self, concentration, index_by_basis):
        self.concentration = concentration
        self.index_by_basis = index_by_basis

    def __call__(self, atoms):
        num_in_template = sum(len(x) for x in self.index_by_basis)
        num_in_atoms = len(atoms)
        ratio = num_in_atoms/num_in_template
        nib = [len(x)*ratio for x in self.index_by_basis]

        if not np.allclose(nib, np.round(nib)):
            return False
        valid = True
        conc = self.concentration
        try:
            x = conc.get_random_concentration(nib=nib)
            x_int = conc.conc_in_int(nib, x)
            x_from_int = conc.to_float_conc(nib, x_int)
            if not conc.is_valid_conc(x_from_int):
                return False
        except Exception:
            valid = False
        return valid


class DistanceBetweenFacetsFilter(CellFilter):
    def __init__(self, ratio):
        self.ratio = ratio

    def _distance_between_facets(self, cell, span):
        v1 = cell[span[0], :]
        v2 = cell[span[1], :]
        normal = np.cross(v1, v2)
        normal /= np.sqrt(normal.dot(normal))

        third_vec = set([0, 1, 2]) - set(span)
        third_vec = list(third_vec)[0]
        d = normal.dot(cell[third_vec, :])
        return abs(d)

    def __call__(self, cell):
        dists = []
        for span in combinations([0, 1, 2], 2):
            dists.append(self._distance_between_facets(cell, span))

        d_min = min(dists)
        d_max = max(dists)
        return d_max/d_min < self.ratio


class VolumeToSurfaceRatioFilter(CellFilter):
    def __init__(self, ratio):
        self.ratio = ratio

    def __call__(self, cell):
        vol = np.linalg.det(cell)
        surf = 0.0
        for span in combinations([0, 1, 2], r=2):
            v1 = cell[span[0], :]
            v2 = cell[span[1], :]
            normal = np.cross(v1, v2)
            area = np.abs(normal.dot(normal))
            surf += 2*area

        factor = area/(6.0*vol**(2.0/3.0))
        return factor < self.ratio


class AngleFilter(CellFilter):
    def __init__(self, min_angle, max_angle):
        self.min_angle = min_angle
        self.max_angle = max_angle

    def __call__(self, cell):
        cos_max = np.cos(self.max_angle*np.pi/180.0)
        cos_min = np.cos(self.min_angle*np.pi/180.0)

        cos_a = []
        for vec in combinations([0, 1, 2], r=2):
            cos_a.append(cell[vec[0], :].dot(cell[vec[1], :]))

        max_ok = all(x > cos_max for x in cos_a)
        min_ok = all(x < cos_min for x in cos_a)
        return max_ok and min_ok



from clease.montecarlo.observers import MCObserver
from ase.io.trajectory import TrajectoryWriter


class Snapshot(MCObserver):
    """Store a snapshot in a trajectory file.

    Parameters:

    fname: str
        Name of the trajectory file (without .traj extension)

    atoms: Atoms
        Instance of the atoms objected modofied by the MC object
    """

    def __init__(self, fname="snapshot", atoms=None):
        super(Snapshot, self).__init__()
        self.name = "Snapshot"
        full_fname = fname + '.traj'
        if atoms is None:
            raise ValueError("No atoms object given!")
        self.atoms = atoms
        self.traj = TrajectoryWriter(full_fname, mode="w")
        self.fname = full_fname

    def __call__(self, system_changes):
        """Write a snapshot to a .traj file.

        Parameters:

        system_changes: list
            System changes. See doc-string of
            `clease.montecarlo.observers.MCObserver`
        """
        self.traj.write(self.atoms)

    def reset(self):
        self.traj = TrajectoryWriter(self.fname, mode="w")
